//*************************************************************************************
// MaxiTweety.c ... Firmware for MaxiTweety Prototype
// Author:  Herbert Burger
// Version: 1.0 / 20.11.2012 / working
// COMPILE WITH -O0 (NO OPTIMIZATION!)
//*************************************************************************************
// Differences to Version 1.0:
// a) "Level" is a moving average to avoid pointer vanishing
//    "Level = (ADCH + Level)/2" replaces "Level = ADCH"
// b) Improved BarGraph algorithm avoids pointer vanishing
// c) New Method of capacitance/inductance measurement without using a relay
//    Stray inductance compensation is no longer required
// d) Option: Compensation of level decay with rising frequency
// *************************************************************************************

// User-definable Options
 
#define FqCorr 931/997		// Frequency Correction Factor:
			// FqCorr = (CpuClock * CorrectFrequency)/(16MHz * DisplayedFrequency)
			// Numerator and denominator must be integer and should be 1000 maximum.
			// Start with CpuClock/16MHz, e.g. 15/16 when CpuClock = 15MHz .
			// If displayed frequency is incorrect then use the long formula.
		
#define Ck 820.				// Calibration Capacitor: xx pF * 100 (type: float !)

#define BG_WIDTH 8			// 8  ... Bargraph Width 8 Character
							// 16 ... Bargraph Width 16 Character
//*************************************************************************************

#define t4ms 250			// Timer2 Compare Value for 4ms TICK

#include <stdint.h>
#include <avr/interrupt.h>

// Global Variables
volatile long Fq, FqOld;	// Measured Frequency
volatile unsigned int Tim0_MSB;
volatile unsigned char Tim1_MSB, Level, Range=0,
	TickCnt, TickAction;	// TickCnt: Count Down Ticks
							// TickAction: 0 ... Do Nothing & Wait,
							// 1 ... Stop & Read Timer0,
							// 2 ... Start Timer0 Frequency Measurement
							// 3 ... Stop & Read Timer1
							// 4 ... Start Timer1 Frequency Measurement
							// 5 ... Auto-Ranging ... Stop & Read Timer1
							// 6 ... Start Timer1 Auto-Ranging Measurement
							// 254 ... Auto-Range Data Ready 
							// 255 ... Frequency & Level Data Ready
							
char kHz[5]="kHz\0", pF[3]="pF\0", uH[3]="uH\0";

float Fo, Fk, Co, Lo;	// for Calculation

long C0, L0;				// for Display
							
//*************************************************************************
// Programmed Delays
//*************************************************************************

void Delay100u(void)			// Wait 100us (@16MHz)
{
	char i; for (i=0;i<155;i++);
}

void Delay4m(void)
{
	char i; for (i=0;i<40;i++) Delay100u();
}

void Delay5m(void)
{
	char i; for (i=0;i<50;i++) Delay100u();
}

void Delay8m(void)
{
	char i; for (i=0;i<80;i++) Delay100u();
}


void Delay100m(void)
{
	char i; for (i=0;i<25;i++) Delay4m();
}


//*************************************************************************
// LCD Configuration 2 / Write Functions
// E = PD7, RS = PC4, DB7 = PC3, DB6 = PC2, DB5 = PC1, DB4 = PC0
//*************************************************************************

void LCD_Pulse(void)
{
	PORTD |= 0x80;			// E=1
	asm volatile ("nop");	// Wait 4 Cycles (4*62,5ns = 250ns)
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	PORTD &= 0x7f;			// E=0
}

void LCD_Write_Nibble(char x)
{
	PORTC = x; LCD_Pulse();
}

void LCD_Cmd(char Cmd)
{
	PORTC = (Cmd>>4)&0x0f; LCD_Pulse();	// High Nibble, RS=0
	PORTC = Cmd&0x0f; LCD_Pulse();		// Low Nibble,  RS=0
	Delay100u();						// "wait >37us"
}

void LCD_Data(char Data)
{
	PORTC = ((Data>>4)&0x0f)|0x10; LCD_Pulse();	// High Nibble, RS=1
	PORTC = (Data&0x0f)|0x10; LCD_Pulse();		// Low Nibble,  RS=1
	Delay100u();								// "wait >(37+4)us"
}

//*************************************************************************
// Basic LCD Functions
//*************************************************************************

void LCD_Clear(void)
{
	LCD_Cmd(0x01);				// Clear Display
	Delay4m();					// wait >1,53ms
}

void LCD_Init(void)
{
	Delay100m();				// wait >40ms
	LCD_Write_Nibble(0x03);
	Delay5m();					// wait >4.1ms
	LCD_Write_Nibble(0x03);
	Delay4m();					// wait > 100us
	LCD_Write_Nibble(0x03);
	Delay100u();				// standard delay (?)
	LCD_Write_Nibble(0x02);		// 4 Bit Interface
	Delay100u();
	LCD_Cmd(0x28);				// 2-Line-Mode, 5x8 Dots
	LCD_Cmd(0x0c);				// Display On, Cursor Off, Blink Off
	LCD_Clear();				// Clear Display (incl. "Home")
	LCD_Cmd(0x06);				// Entry Mode: Increment Address / No Shift
}
	
void LCD_Pointer(char Zeile, char Spalte){
	LCD_Cmd(--Zeile*64 + --Spalte + 128);// Calculate Text Buffer Address
}

void LCD_String(char *String){
	while(*String)				// Output Text Until End-Of-String(0x00)
	{
		LCD_Data(*String); String++;
	}
}

//*************************************************************************
// LCD Pseudo Graphics (Pointer Instrument Like)
//*************************************************************************

void LCD_BarGraphChar(void)					//Define Special Character Set
{
	char i,k;
    char a[6]={0x00,0x10,0x08,0x04,0x02,0x01};	//Microline Templates
	LCD_Cmd(0x40);								//CGRAM Pointer = 0
	for (i=0;i<6;i++)							//for Each Special Character
	{
		LCD_Data(0x1f);							//Top Bar
		for (k=0;k<5;k++) LCD_Data(a[i]);		//5* the Microline
		LCD_Data(0x1f);							//Botom Bar
		LCD_Data(0);							//8th Microline empty
	}
}
	
#if BG_WIDTH==16
void LCD_BarGraph(unsigned char x) 				//Generate Bargraph
{
	char i=16;							//Character Counter
	while (x>15)						//Full Rectangles
		{
		LCD_Data(0);
		x -= 16; i--;
		}
	if (x==0) x=1;						//avoid pointer vanishing
	LCD_Data((x+2)/3); i--;				//Micro Columns
	while (i--) LCD_Data(0);			//Clear Display to the Right
}
#endif

#if BG_WIDTH==8
void LCD_BarGraph(unsigned char x) 				//Generate Bargraph
{
	char i=8;							//Character Counter
	while (x>31)						//Full Rectangles
		{
		LCD_Data(0);
		x -= 32; i--;
		}
	if (x==0) x=1; if (x==31) x=30;		//avoid pointer vanishing
	LCD_Data((x+5)/6); i--;				//Micro Columns
while (i--) LCD_Data(0);				//Clear Display to the Right
}
#endif

//*************************************************************************************
// Frequency Measurement
//*************************************************************************************

ISR(TIMER0_OVF_vect)		// Overflow Counter for Timer0 (16Bit)
{
	Tim0_MSB++;
}

ISR(TIMER1_OVF_vect)		// Overflow Counter for Timer1 (8Bit)
{
	Tim1_MSB++;
}


ISR(TIMER2_COMP_vect)		// 4ms-TICK: Timer2 Compare Match Interrupt
{
	if (--TickCnt == 0)
	{
		switch (TickAction)
		{
			case 0: TickCnt = 1; break;	// Do Nothing & Wait			 
			case 1: TCCR0 = 0;			// Stop & Read Timer0
				Fq = (long) 256*Tim0_MSB + TCNT0; TickAction = 255;
				Level = (ADCH + Level)/2; TickCnt = 1; break;
			case 2: TCNT0 = 0;			// Start Timer0 Frequency Measurement
				Tim0_MSB = 0; TCCR0 = 7; TickCnt = 25; TickAction = 1; break;
			case 3: TCCR1B = 0;			// Stop & Read Timer1
				Fq = (long) 65536*Tim1_MSB + TCNT1; TickAction = 255;
				Level = (ADCH + Level)/2; TickCnt = 1; break;
			case 4: TCNT1 = 0;			// Start Timer1 Frequency Measurement
				TCCR1B = 7; Tim1_MSB = 0; TickCnt = 25; TickAction = 3; break;
			case 5: TCCR1B = 0;			// Auto-Ranging ... Stop & Read Timer1
				Fq = (long) TCNT1; TickAction = 254;
				TickCnt = 1; break;
			case 6: TCNT1 = 0;			// Start Timer1 Auto-Ranging Measurement
				TCCR1B = 7; TickAction = 5;
			default: TickCnt = 1;
		 }
	}	
}

void FreqCorr(void)
{
	Fq = Fq * FqCorr;
}

void AutoRange(void)
{
	char RangeOK=0;
	unsigned int RangeLimit[6]={0,500,5000,10000,20000,64000};
	
	DDRB = 4; PORTB = 4; //Prescaler: / 40
	TickAction = 6;		// Start Auto Range
	do { asm volatile ("nop"); } while (TickAction!=254);
	FreqCorr();
	
	switch (Range)
	{
		case 1: if (Fq<525) RangeOK=1; break;
		case 2: if ((Fq>475)&&(Fq<5250)) RangeOK=1; break;
		case 3: if ((Fq>4750)&&(Fq<10500)) RangeOK=1; break;
		case 4: if ((Fq>9500)&&(Fq<21000)) RangeOK=1; break;
		case 5: if (Fq>19000) RangeOK=1;		 
	}
	if (RangeOK!=1)
	{
		for(Range=1;Range<6;Range++) if ((Fq>=RangeLimit[Range-1])&&(Fq<RangeLimit[Range])) break;
	}	
}

void GetFrequency(void)
{
	
	char PreScale[6]={0,0,7,6,4,0}, Actions[6]={0,2,4,4,4,4};
	int MultiPly[6]={1,1,10,20,40,80};	// This Gives 1/10 of the Real Frequency!!!
	int Mult;

// Switch to Range	
	DDRB = PreScale[Range]; PORTB = PreScale[Range]; Mult = MultiPly[Range]; TickAction = Actions[Range];
		
// Read Result
	do { asm volatile ("nop"); } while (TickAction!=255) ;
	FreqCorr(); Fq *= Mult;
}

void DisplayNum(int n, char c)
{
	if (c==3) goto D_100;	// c==3: 3 Digits, No Zero Suppression
	if (c==2) goto D_10;	// c==2: 2 Digits, No Zero Suppression
	if (n<10) goto D_1;		// c==0: max 3 Digits, Zero suppression 
	if (n<100) goto D_10;
	D_100: c = n/100; n %= 100; LCD_Data(c+48);
	D_10: c = n/10; n %= 10; LCD_Data(c+48);
	D_1: LCD_Data(n+48);	
}

void DisplayNumber(long x)
{
	int n;
	if (x<100000) goto x_100k;
	n = x/100000; x %= 100000; DisplayNum(n,0); LCD_Data('.');
x_100k:	n = x/100; DisplayNum(n,3); LCD_Data(',');
	n = x%100; DisplayNum(n,2);
}
//*************************************************************************
// Main
//*************************************************************************

int main(void)
{
	
	PORTC = 0;				// PC6 = /RESET ... Input, Pullup = Off
	DDRC = 0b00011111;		// PC5 = AIN5   ... Input, Pullup = Off
							// PC4 - PC0 Output
	PORTD = 0b01001000;		// PD7 = E = 0  ... Output, PD6 = 1 ... Relay Driver (Off)
	DDRD = 0b11000010;		// PD7, PD6, PD1 .. Output, PD3 ... Input with Pull-Up Resistor

// init Timer0
	TCCR0 = 0;				// Timer0: 8-Bit Counter, STOP = 0 / RUN = 7
	TCNT0 = 0;
	
// init Timer1
	TCCR1A = 0;				// Timer1: 16-Bit Counter,
	TCCR1B = 0;				// STOP = 0 / RUN = 7
	
// init Timer2 (4ms TICK)
	TCCR2 = 0x0E;			// Timer2: Clear on Terminal Count,
	OCR2 = 250;				// Output OC2 not used, Prescaler: /256

// Timer Interrupts
	TIMSK = 0x85;			// Timer Interrupt Mask:
							// Timer0: Overflow (TIMER0 OVF, Vector 9)
							// Timer1: Overflow (TIMER1 OVF, Vector 8)
							// Timer2: Compare  (TIMER2 COMP, Vector 3)

// init ADC
	ADMUX = 0xE5;			// ADC: Internal Reference, Input on Pin PC5,
							// left bounded,
	ADCSRA = 0xE7;			// 9,6 kHz Free Run, START, no Interrupts
	
// init LCD	
	LCD_Init();
	LCD_BarGraphChar();

// Start TICK State Machine
	TickAction = 0;			// TICK ISR does nothing
	TickCnt = 1;
	sei();					// Global Interrupt Enable

//****************************************************************************
// Main Loop
//****************************************************************************
    while(1)
    {
//		
		
		FqOld = Fq;								// Save previous frequency result
		AutoRange();
		GetFrequency();
		if ((PIND&0b00000100)==0)				// L/C measurement?
		{
			Fo = (float) FqOld;					// Frequency Without Calibration Capacitor
												// Discard latest frequency result
			Delay100m();						// Wait 100ms
			GetFrequency(); Fk = (float) Fq;	// Frequency With Calibration Capacitor
			Fk = Fo/Fk; Co = Ck/(Fk*Fk-1);
			Lo = 2.53303E18F / (Fo*Fo*Co);
			L0 = (long) Lo; C0 = (long) Co;		// These are the LC Results
			LCD_Pointer(1,1);					// Display Inductance
			DisplayNumber(L0); LCD_String(uH);
			LCD_Pointer(2,1);					// Display Capacitance
			DisplayNumber(C0); LCD_String(pF);
			//Delay100m();
			while((PIND&0b00000100)==0);		// Wait for LC Key Released
		}
		else
		{
			LCD_Pointer(1,1);		// Display Frequency
			DisplayNumber(Fq); LCD_String(kHz);
			LCD_Pointer(2,1); LCD_BarGraph(Level); // Display Level
		}		
	}
}